/*
 *
 *  BlueZ - Bluetooth protocol stack for Linux
 *
 *  Copyright (C) 2005-2007  Marcel Holtmann <marcel@holtmann.org>
 *
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include <dbus/dbus-glib.h>

#ifdef HAVE_HAL
#include <dbus/dbus-glib-lowlevel.h>
#include <hal/libhal.h>
#endif

#include <gconf/gconf-client.h>

#include <glib/gi18n.h>

#include <gtk/gtk.h>

#include "client.h"

static BluetoothClient *client;

static DBusGConnection *conn;

typedef enum {
	ICON_POLICY_NEVER,
	ICON_POLICY_ALWAYS,
	ICON_POLICY_PRESENT,
} EnumIconPolicy;

static int icon_policy = ICON_POLICY_PRESENT;

#define PREF_DIR		"/apps/bluetooth-manager"
#define PREF_USE_HAL		PREF_DIR "/use_hal"
#define PREF_ICON_POLICY	PREF_DIR "/icon_policy"
#define PREF_AUTO_AUTHORIZE	PREF_DIR "/auto_authorize"

static GConfEnumStringPair icon_policy_enum_map [] = {
	{ ICON_POLICY_NEVER,	"never"		},
	{ ICON_POLICY_ALWAYS,	"always"	},
	{ ICON_POLICY_PRESENT,	"present"	},
	{ ICON_POLICY_PRESENT,	NULL		},
};

static GConfClient* gconf;

static GtkWidget *notebook;
static GtkWidget *button_properties;
static GtkWidget *button_autoauth;
static GtkWidget *button_usehal;
static GtkWidget *button_never;
static GtkWidget *button_always;
static GtkWidget *button_present;
static GtkWidget *device_tree;

static GtkListStore *service_store;

static void update_icon_policy(GtkWidget *button)
{
	GtkWidget *widget;
	const char *str;

	if (button == NULL) {
		switch (icon_policy) {
		case ICON_POLICY_NEVER:
			widget = button_never;
			break;
		case ICON_POLICY_ALWAYS:
			widget = button_always;
			break;
		case ICON_POLICY_PRESENT:
		default:
			widget = button_present;
			break;
		}

		if (!widget)
			return;

		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget), TRUE);

		return;
	}

	if (button == button_always)
		icon_policy = ICON_POLICY_ALWAYS;
	else if (button == button_never)
		icon_policy = ICON_POLICY_NEVER;
	else if (button == button_present)
		icon_policy = ICON_POLICY_PRESENT;

	str = gconf_enum_to_string(icon_policy_enum_map, icon_policy);

	gconf_client_set_string(gconf, PREF_ICON_POLICY, str, NULL);
}

static void delete_callback(GtkWidget *window, GdkEvent *event,
						gpointer user_data)
{
	gtk_widget_destroy(GTK_WIDGET(window));

	gtk_main_quit();
}

static void close_callback(GtkWidget *button, gpointer user_data)
{
	GtkWidget *window = user_data;

	gtk_widget_destroy(GTK_WIDGET(window));

	gtk_main_quit();
}

static void policy_callback(GtkWidget *button, gpointer user_data)
{
	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button)) == TRUE)
		update_icon_policy(button);
}

static void autoauth_callback(GtkWidget *button, gpointer user_data)
{
	gboolean value;

	value = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button));

	gconf_client_set_bool(gconf, PREF_AUTO_AUTHORIZE, value, NULL);
}

#ifdef HAVE_HAL
static void usehal_callback(GtkWidget *button, gpointer user_data)
{
	gboolean value;

	value = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button));

	gconf_client_set_bool(gconf, PREF_USE_HAL, value, NULL);
}
#endif

static GtkWidget *create_label(const gchar *str)
{
	GtkWidget *label;
	gchar *tmp;

	label = gtk_label_new(NULL);

	tmp = g_strdup_printf("<b>%s</b>", str);

	gtk_label_set_markup(GTK_LABEL(label), tmp);

	g_free(tmp);

	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);

	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);

	return label;
}

static void create_general(void)
{
	GtkWidget *vbox;
	GtkWidget *label;
	GSList *group = NULL;
	gboolean value;

	vbox = gtk_vbox_new(FALSE, 6);

	gtk_container_set_border_width(GTK_CONTAINER(vbox), 12);

	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), vbox, NULL);

	gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), vbox, _("General"));

	label = create_label(_("Authorization requests"));

	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	button_autoauth = gtk_check_button_new_with_label(
			_("Automatically authorize incoming requests"));

	value = gconf_client_get_bool(gconf, PREF_AUTO_AUTHORIZE, NULL);

	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button_autoauth), value);

	gtk_box_pack_start(GTK_BOX(vbox), button_autoauth, FALSE, FALSE, 0);

	g_signal_connect(G_OBJECT(button_autoauth), "toggled",
					G_CALLBACK(autoauth_callback), NULL);

#ifdef HAVE_HAL
	label = create_label(_("\nHardware database"));

	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	button_usehal = gtk_check_button_new_with_label(
				_("Select class of device automatically"));

	value = gconf_client_get_bool(gconf, PREF_USE_HAL, NULL);

	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button_usehal), value);

	gtk_box_pack_start(GTK_BOX(vbox), button_usehal, FALSE, FALSE, 0);

	g_signal_connect(G_OBJECT(button_usehal), "toggled",
					G_CALLBACK(usehal_callback), NULL);
#endif

	label = create_label(_("\nNotification area"));

	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	button_never = gtk_radio_button_new_with_label(group,
						_("Never display icon"));

	group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(button_never));

	gtk_box_pack_start(GTK_BOX(vbox), button_never, FALSE, FALSE, 0);

	g_signal_connect(G_OBJECT(button_never), "toggled",
					G_CALLBACK(policy_callback), NULL);

	button_present = gtk_radio_button_new_with_label(group,
				_("Only display when adapter present"));

	group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(button_present));

	gtk_box_pack_start(GTK_BOX(vbox), button_present, FALSE, FALSE, 0);

	g_signal_connect(G_OBJECT(button_present), "toggled",
					G_CALLBACK(policy_callback), NULL);

	button_always = gtk_radio_button_new_with_label(group,
						_("Always display icon"));

	group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(button_always));

	gtk_box_pack_start(GTK_BOX(vbox), button_always, FALSE, FALSE, 0);

	g_signal_connect(G_OBJECT(button_always), "toggled",
					G_CALLBACK(policy_callback), NULL);

	update_icon_policy(NULL);
}

#if 0
static void startstop_notify(DBusGProxy *object,
				DBusGProxyCall *call, void *user_data)
{
	gchar *path = user_data;
	GtkTreeIter iter;
	gboolean active;

	dbus_g_proxy_end_call(object, call, NULL, G_TYPE_INVALID);

	if (gtk_tree_model_get_iter_from_string(GTK_TREE_MODEL(service_store),
							&iter, path) == FALSE)
		return;

	dbus_g_proxy_call(object, "IsRunning", NULL, G_TYPE_INVALID,
				G_TYPE_BOOLEAN, &active, G_TYPE_INVALID);

	gtk_list_store_set(service_store, &iter, 2, active, 3, FALSE, -1);
}

static void startstop_callback(GtkCellRendererToggle *renderer,
					gchar *path, gpointer user_data)
{
	DBusGProxy *object;
	GtkTreeIter iter;
	const char *service_path;
	gboolean active;

	if (gtk_tree_model_get_iter_from_string(GTK_TREE_MODEL(service_store),
							&iter, path) == FALSE)
		return;

	gtk_tree_model_get(GTK_TREE_MODEL(service_store),
				&iter, 0, &service_path, 2, &active, -1);

	object = dbus_g_proxy_new_for_name(conn, "org.bluez",
					service_path, "org.bluez.Service");

	gtk_list_store_set(service_store, &iter, 3, TRUE, -1);

	if (active == FALSE) {
		dbus_g_proxy_begin_call(object, "Start", startstop_notify,
				g_strdup(path), g_free, G_TYPE_INVALID);
		active = TRUE;
	} else {
		dbus_g_proxy_begin_call(object, "Stop", startstop_notify,
				g_strdup(path), g_free, G_TYPE_INVALID);
		active = FALSE;
	}
}

static void status_to_toggle(GtkTreeViewColumn *column, GtkCellRenderer *cell,
			GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	gboolean active, progress;

	gtk_tree_model_get(model, iter, 2, &active, 3, &progress, -1);

	g_object_set(cell, "active", active, NULL);
	g_object_set(cell, "sensitive", progress == TRUE ? FALSE : TRUE, NULL);
}

static void status_to_text(GtkTreeViewColumn *column, GtkCellRenderer *cell,
			GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	gboolean active, progress;
	gchar *str;

	gtk_tree_model_get(model, iter, 2, &active, 3, &progress, -1);

	if (active == TRUE) {
		if (progress == TRUE)
			str = "Stopping";
		else
			str = "Running";
	} else {
		if (progress == TRUE)
			str = "Starting";
		else
			str = "Stopped";
	}
 
	g_object_set(cell, "text", str, NULL);
}

static void show_properties(const gchar *path, const gchar *description)
{
	GtkWidget *dialog;
	GtkWidget *button;
	GtkWidget *vbox;
	GtkWidget *label;
	gint response;

	dialog = gtk_dialog_new();

	gtk_window_set_title(GTK_WINDOW(dialog), _("Service properties"));

	gtk_window_set_icon_name(GTK_WINDOW(dialog), "stock_bluetooth");

	gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER_ON_PARENT);

	gtk_window_set_default_size(GTK_WINDOW(dialog), 240, -1);

	gtk_dialog_set_has_separator(GTK_DIALOG(dialog), FALSE);

	button = gtk_dialog_add_button(GTK_DIALOG(dialog),
					GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE);

	gtk_widget_grab_default(button);

	vbox = gtk_vbox_new(FALSE, 6);

	gtk_container_set_border_width(GTK_CONTAINER(vbox), 12);

	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), vbox);

	label = create_label(_("Service path"));

	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	label = gtk_label_new(path);

	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);

	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);

	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	label = create_label(_("\nDescription"));

	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	label = gtk_label_new(description);

	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);

	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.0);

	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	gtk_widget_show_all(GTK_WIDGET(dialog));

	response = gtk_dialog_run(GTK_DIALOG(dialog));

	gtk_widget_destroy(GTK_WIDGET(dialog));
}

static void properties_callback(GtkWidget *button, gpointer user_data)
{
	GtkTreeSelection *selection = user_data;
	GtkTreeIter iter;
	GtkTreeModel *model;
	gchar *path, *description;

	if (gtk_tree_selection_get_selected(selection, &model, &iter) == TRUE) {
		gtk_tree_model_get(model, &iter, 0, &path, 4, &description, -1);

		show_properties(path, description);

		g_free(path);
	}
}

static void select_callback(GtkTreeSelection *selection, gpointer user_data)
{
	gboolean sensitive;

	sensitive = gtk_tree_selection_get_selected(selection, NULL, NULL);

	gtk_widget_set_sensitive(GTK_WIDGET(button_properties), sensitive);
}

static void create_services(void)
{
	GtkWidget *vbox;
	GtkWidget *label;
	GtkWidget *scrolled;
	GtkWidget *tree;
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
	GtkTreeSelection *selection;
	GtkWidget *buttonbox;

	vbox = gtk_vbox_new(FALSE, 6);

	gtk_container_set_border_width(GTK_CONTAINER(vbox), 12);

	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), vbox, NULL);

	gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), vbox, _("Services"));

	label = create_label(_("Available services"));

	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	scrolled = gtk_scrolled_window_new(NULL, NULL);

	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
				GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolled),
							GTK_SHADOW_OUT);

	gtk_container_add(GTK_CONTAINER(vbox), scrolled);

	tree = gtk_tree_view_new();

	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(tree), TRUE);

	gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(tree), TRUE);

	gtk_tree_view_set_model(GTK_TREE_VIEW(tree),
					GTK_TREE_MODEL(service_store));

	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), 0,
					"Name", gtk_cell_renderer_text_new(),
							"text", 1, NULL);

	column = gtk_tree_view_get_column(GTK_TREE_VIEW(tree), 0);

	gtk_tree_view_column_set_expand(GTK_TREE_VIEW_COLUMN(column), TRUE);

	column = gtk_tree_view_column_new();

	gtk_tree_view_column_set_title(column, "Status");

	gtk_tree_view_column_set_spacing(column, 2);

	renderer = gtk_cell_renderer_toggle_new();
	gtk_tree_view_column_pack_start(column, renderer, FALSE);
	gtk_tree_view_column_set_cell_data_func(column, renderer,
						status_to_toggle, NULL, NULL);

	g_signal_connect(G_OBJECT(renderer), "toggled",
				G_CALLBACK(startstop_callback), NULL);

	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(column, renderer, TRUE);
	gtk_tree_view_column_set_cell_data_func(column, renderer,
						status_to_text, NULL, NULL);

	gtk_tree_view_append_column(GTK_TREE_VIEW(tree), column);

	gtk_tree_view_column_set_min_width(GTK_TREE_VIEW_COLUMN(column), 100);
	gtk_tree_view_column_set_max_width(GTK_TREE_VIEW_COLUMN(column), 100);

	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));

	gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);

	g_signal_connect(G_OBJECT(selection), "changed",
				G_CALLBACK(select_callback), NULL);

	gtk_container_add(GTK_CONTAINER(scrolled), tree);

	buttonbox = gtk_hbutton_box_new();

	gtk_button_box_set_layout(GTK_BUTTON_BOX(buttonbox), GTK_BUTTONBOX_START);

	gtk_box_pack_start(GTK_BOX(vbox), buttonbox, FALSE, FALSE, 0);

	button_properties = gtk_button_new_from_stock(GTK_STOCK_PROPERTIES);

	gtk_widget_set_sensitive(GTK_WIDGET(button_properties), FALSE);

	g_signal_connect(G_OBJECT(button_properties), "clicked",
				G_CALLBACK(properties_callback), selection);

	gtk_container_add(GTK_CONTAINER(buttonbox), button_properties);
}

static gboolean selection_func(GtkTreeSelection *selection,
				GtkTreeModel *model, GtkTreePath *path,
				gboolean currently_selected, gpointer data)
{
	GtkTreeIter iter;
	const gchar *value;

	gtk_tree_model_get_iter(model, &iter, path);

	gtk_tree_model_get(model, &iter, COLUMN_PATH, &value, -1);

	return value ? FALSE : TRUE;
}

static void connected_to_icon(GtkTreeViewColumn *column, GtkCellRenderer *cell,
			GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	gboolean connected;

	gtk_tree_model_get(model, iter, COLUMN_CONNECTED, &connected, -1);

	if (connected == TRUE)
		g_object_set(cell, "icon-name", GTK_STOCK_CONNECT, NULL);

	g_object_set(cell, "visible", connected, NULL);
}

static void trusted_to_icon(GtkTreeViewColumn *column, GtkCellRenderer *cell,
			GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	gboolean trusted;

	gtk_tree_model_get(model, iter, COLUMN_TRUSTED, &trusted, -1);

	if (trusted == TRUE)
		g_object_set(cell, "icon-name",
				GTK_STOCK_DIALOG_AUTHENTICATION, NULL);

	g_object_set(cell, "visible", trusted, NULL);
}

static void create_devices(void)
{
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *label;
	GtkWidget *scrolled;
	GtkWidget *tree;
	GtkTreeModel *model;
	GtkTreeViewColumn *column;
	GtkTreeSelection *selection;
	GtkWidget *button;
	GtkWidget *buttonbox;

	vbox = gtk_vbox_new(FALSE, 6);

	gtk_container_set_border_width(GTK_CONTAINER(vbox), 12);

	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), vbox, NULL);

	gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), vbox, _("Devices"));

	label = create_label(_("Trusted devices"));

	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	hbox = gtk_hbox_new(FALSE, 6);

	gtk_container_set_border_width(GTK_CONTAINER(vbox), 12);

	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);

	scrolled = gtk_scrolled_window_new(NULL, NULL);

	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
				GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolled),
							GTK_SHADOW_OUT);

	gtk_container_add(GTK_CONTAINER(hbox), scrolled);

	tree = gtk_tree_view_new();

	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(tree), FALSE);

	gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(tree), TRUE);

	model = bluetooth_client_get_model(client);

	gtk_tree_view_set_model(GTK_TREE_VIEW(tree), model);

	g_object_unref(model);

	gtk_tree_view_expand_all(GTK_TREE_VIEW(tree));

	gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(tree), 0,
					"Device", gtk_cell_renderer_text_new(),
							"text", COLUMN_NAME, NULL);

	column = gtk_tree_view_get_column(GTK_TREE_VIEW(tree), 0);

	gtk_tree_view_column_set_max_width(GTK_TREE_VIEW_COLUMN(column), 50);

	gtk_tree_view_column_set_expand(GTK_TREE_VIEW_COLUMN(column), TRUE);

	gtk_tree_view_insert_column_with_data_func(GTK_TREE_VIEW(tree), -1,
					"Connected", gtk_cell_renderer_pixbuf_new(),
					connected_to_icon, NULL, NULL);

	gtk_tree_view_insert_column_with_data_func(GTK_TREE_VIEW(tree), -1,
					"Trusted", gtk_cell_renderer_pixbuf_new(),
					trusted_to_icon, NULL, NULL);

	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree));

	gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);

	gtk_tree_selection_set_select_function(selection,
						selection_func, NULL, NULL);

	//g_signal_connect(G_OBJECT(selection), "changed",
	//			G_CALLBACK(select_callback), NULL);

	gtk_container_add(GTK_CONTAINER(scrolled), tree);

	device_tree = tree;

	buttonbox = gtk_vbutton_box_new();

	gtk_box_set_spacing(GTK_BOX(buttonbox), 6);

	gtk_button_box_set_layout(GTK_BUTTON_BOX(buttonbox), GTK_BUTTONBOX_START);

	gtk_box_pack_start(GTK_BOX(hbox), buttonbox, FALSE, FALSE, 0);

	button = gtk_button_new_from_stock(GTK_STOCK_DELETE);

	//gtk_widget_set_sensitive(GTK_WIDGET(button), FALSE);

	gtk_container_add(GTK_CONTAINER(buttonbox), button);

	button = gtk_button_new_from_stock(GTK_STOCK_DISCONNECT);

	//gtk_widget_set_sensitive(GTK_WIDGET(button), FALSE);

	//g_signal_connect(G_OBJECT(button), "clicked",
	//			G_CALLBACK(properties_callback), selection);

	gtk_container_add(GTK_CONTAINER(buttonbox), button);
}
#endif

static void create_window(void)
{
	GtkWidget *window;
	GtkWidget *vbox;
	GtkWidget *buttonbox;
	GtkWidget *button;

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

	gtk_window_set_title(GTK_WINDOW(window), _("Bluetooth preferences"));

	gtk_window_set_icon_name(GTK_WINDOW(window), "stock_bluetooth");

	gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);

	gtk_window_set_default_size(GTK_WINDOW(window), 380, 440);

	g_signal_connect(G_OBJECT(window), "delete-event",
					G_CALLBACK(delete_callback), NULL);

	vbox = gtk_vbox_new(FALSE, 12);

	gtk_container_set_border_width(GTK_CONTAINER(vbox), 12);

	gtk_container_add(GTK_CONTAINER(window), vbox);

	gtk_box_pack_start(GTK_BOX(vbox), notebook, TRUE, TRUE, 0);

	buttonbox = gtk_hbutton_box_new();

	gtk_button_box_set_layout(GTK_BUTTON_BOX(buttonbox), GTK_BUTTONBOX_END);

	gtk_box_pack_start(GTK_BOX(vbox), buttonbox, FALSE, FALSE, 0);

	button = gtk_button_new_from_stock(GTK_STOCK_CLOSE);

	g_signal_connect(G_OBJECT(button), "clicked",
					G_CALLBACK(close_callback), window);

	gtk_container_add(GTK_CONTAINER(buttonbox), button);

#if 0
	button = gtk_button_new_from_stock(GTK_STOCK_HELP);

	gtk_container_add(GTK_CONTAINER(buttonbox), button);

	gtk_button_box_set_child_secondary(GTK_BUTTON_BOX(buttonbox),
								button, TRUE);
#endif

#if 0
	create_devices();

	create_services();
#endif

	create_general();

	gtk_widget_show_all(window);
}

static GList *adapter_list = NULL;

struct adapter_data {
	char *path;
	int attached;
	GtkWidget *child;
	GtkWidget *button_connect;
	GtkWidget *button_visible;
	GtkWidget *button_limited;
	GtkWidget *timeout_label;
	GtkWidget *timeout_scale;
	GtkWidget *entry;
	GtkWidget *combo;
	int name_changed;
	gboolean enable_limited;
};

static void adapter_free(gpointer data, gpointer user_data)
{
	struct adapter_data *adapter = data;

	adapter_list = g_list_remove(adapter_list, adapter);

	g_free(adapter->path);
	g_free(adapter);
}

static void adapter_disable(gpointer data, gpointer user_data)
{
	struct adapter_data *adapter = data;
	gint page;

	adapter->attached = 0;

	if (adapter->child != NULL) {
		page = gtk_notebook_page_num(GTK_NOTEBOOK(notebook),
							adapter->child);

		gtk_notebook_remove_page(GTK_NOTEBOOK(notebook), page);
	}

	adapter->child = NULL;
}

static gint adapter_compare(gconstpointer a, gconstpointer b)
{
	const struct adapter_data *adapter = a;
	const char *path = b;

	return strcmp(adapter->path, path);
}

static void mode_callback(GtkWidget *button, gpointer user_data)
{
	struct adapter_data *adapter = user_data;
	const char *mode;
	DBusGProxy *object;
	gboolean sensitive;

	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button)) == FALSE)
		return;

	if (button == adapter->button_connect) {
		sensitive = FALSE;
		mode = "connectable";
	} else if (button == adapter->button_visible) {
		sensitive = TRUE;
		mode = "discoverable";
	} else if (button == adapter->button_limited) {
		sensitive = adapter->enable_limited;
		mode = "limited";
	} else
		return;

	object = dbus_g_proxy_new_for_name(conn, "org.bluez",
					adapter->path, "org.bluez.Adapter");

	dbus_g_proxy_call(object, "SetMode", NULL,
			G_TYPE_STRING, mode, G_TYPE_INVALID, G_TYPE_INVALID);

	gtk_widget_set_sensitive(GTK_WIDGET(adapter->timeout_label), sensitive);
	gtk_widget_set_sensitive(GTK_WIDGET(adapter->timeout_scale), sensitive);
}

static void scale_callback(GtkWidget *scale, gpointer user_data)
{
	struct adapter_data *adapter = user_data;
	gdouble value;
	guint32 timeout;
	DBusGProxy *object;

	value = gtk_range_get_value(GTK_RANGE(scale));

	if (value == 31)
		timeout = 0;
	else
		timeout = value * 60;

	object = dbus_g_proxy_new_for_name(conn, "org.bluez",
					adapter->path, "org.bluez.Adapter");

	dbus_g_proxy_call(object, "SetDiscoverableTimeout", NULL,
			G_TYPE_UINT, timeout, G_TYPE_INVALID, G_TYPE_INVALID);
}

static gchar *format_callback(GtkWidget *scale, gdouble value, gpointer user_data)
{
	if (value > 30)
		return g_strdup(_("never"));
	else if (value == 1)
		return g_strdup(_("1 minute"));
	else
		return g_strdup_printf(_("%g minutes"), value);
}

static void name_callback(GtkWidget *editable, gpointer user_data)
{
	struct adapter_data *adapter = user_data;

	adapter->name_changed = 1;
}

static gboolean focus_callback(GtkWidget *editable,
				GdkEventFocus *event, gpointer user_data)
{
	struct adapter_data *adapter = user_data;
	const gchar *text;
	DBusGProxy *object;

	if (!adapter->name_changed)
		return FALSE;

	text = gtk_entry_get_text(GTK_ENTRY(editable));

	object = dbus_g_proxy_new_for_name(conn, "org.bluez",
					adapter->path, "org.bluez.Adapter");

	dbus_g_proxy_call(object, "SetName", NULL,
			G_TYPE_STRING, text, G_TYPE_INVALID, G_TYPE_INVALID);

	adapter->name_changed = 0;

	return FALSE;
}

static void class_callback(GtkWidget *combobox, gpointer user_data)
{
	struct adapter_data *adapter = user_data;
	const char *minor;
	gint index;
	DBusGProxy *object;

	index = gtk_combo_box_get_active(GTK_COMBO_BOX(combobox));

	switch (index) {
	case 0:
		minor = "uncategorized";
		break;
	case 1:
		minor = "desktop";
		break;
	case 2:
		minor = "laptop";
		break;
	default:
		return;
	}

	object = dbus_g_proxy_new_for_name(conn, "org.bluez",
					adapter->path, "org.bluez.Adapter");

	dbus_g_proxy_call(object, "SetMinorClass", NULL,
			G_TYPE_STRING, minor, G_TYPE_INVALID, G_TYPE_INVALID);
}

static void create_adapter(struct adapter_data *adapter)
{
	DBusGProxy *object;
	const char *name = NULL, *mode = NULL, *major = NULL, *minor = NULL;
	const guint32 timeout;
	const gchar **array = NULL;

	GtkWidget *vbox;
	GtkWidget *label;
	GtkWidget *button;
	GtkWidget *scale;
	GtkWidget *entry;
	GtkWidget *combobox;
	GSList *group = NULL;
	gdouble value;
	gint index;

	object = dbus_g_proxy_new_for_name(conn, "org.bluez",
					adapter->path, "org.bluez.Adapter");

	dbus_g_proxy_call(object, "ListAvailableModes", NULL, G_TYPE_INVALID,
				G_TYPE_STRV, &array, G_TYPE_INVALID);

	adapter->enable_limited = FALSE;

	if (array) {
		while (*array) {
			if (!strcmp(*array, "limited"))
				adapter->enable_limited = TRUE;
			array++;
		}
	}

	dbus_g_proxy_call(object, "GetMode", NULL, G_TYPE_INVALID,
				G_TYPE_STRING, &mode, G_TYPE_INVALID);

	dbus_g_proxy_call(object, "GetDiscoverableTimeout", NULL, G_TYPE_INVALID,
				G_TYPE_UINT, &timeout, G_TYPE_INVALID);

	dbus_g_proxy_call(object, "GetName", NULL, G_TYPE_INVALID,
				G_TYPE_STRING, &name, G_TYPE_INVALID);

	dbus_g_proxy_call(object, "GetMajorClass", NULL, G_TYPE_INVALID,
				G_TYPE_STRING, &major, G_TYPE_INVALID);

	dbus_g_proxy_call(object, "GetMinorClass", NULL, G_TYPE_INVALID,
				G_TYPE_STRING, &minor, G_TYPE_INVALID);

	vbox = gtk_vbox_new(FALSE, 6);

	gtk_container_set_border_width(GTK_CONTAINER(vbox), 12);

	gtk_notebook_prepend_page(GTK_NOTEBOOK(notebook), vbox, NULL);

	gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), vbox,
				name && strlen(name) > 0 ? name : _("Adapter"));

	label = gtk_notebook_get_tab_label(GTK_NOTEBOOK(notebook), vbox);

	gtk_label_set_max_width_chars(GTK_LABEL(label), 20);

	gtk_label_set_ellipsize(GTK_LABEL(label), PANGO_ELLIPSIZE_END);

	adapter->child = vbox;

	label = create_label(_("Mode of operation"));

	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	button = gtk_radio_button_new_with_label(group,
					_("Other devices can connect"));

	group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(button));

	gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);

	if (mode && !strcmp(mode, "off"))
		gtk_widget_set_sensitive(GTK_WIDGET(button), FALSE);

	if (mode && !strcmp(mode, "connectable"))
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);

	adapter->button_connect = button;

	g_signal_connect(G_OBJECT(button), "toggled",
					G_CALLBACK(mode_callback), adapter);

	button = gtk_radio_button_new_with_label(group,
				_("Visible and connectable for other devices"));

	group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(button));

	gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);

	if (mode && !strcmp(mode, "off"))
		gtk_widget_set_sensitive(GTK_WIDGET(button), FALSE);

	if (mode && !strcmp(mode, "discoverable"))
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);

	adapter->button_visible = button;

	g_signal_connect(G_OBJECT(button), "toggled",
					G_CALLBACK(mode_callback), adapter);

	button = gtk_radio_button_new_with_label(group,
				_("Limited discoverable and connectable"));

	group = gtk_radio_button_get_group(GTK_RADIO_BUTTON(button));

	gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);

	if (mode && !strcmp(mode, "off"))
		gtk_widget_set_sensitive(GTK_WIDGET(button), FALSE);

	if (mode && !strcmp(mode, "limited"))
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);

	if (adapter->enable_limited == FALSE)
		gtk_widget_set_sensitive(GTK_WIDGET(button), FALSE);

	adapter->button_limited = button;

	g_signal_connect(G_OBJECT(button), "toggled",
					G_CALLBACK(mode_callback), adapter);

	label = gtk_label_new(_("Make adapter invisible after:"));

	gtk_misc_set_alignment(GTK_MISC(label), 0, 0);

	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	adapter->timeout_label = label;

	scale = gtk_hscale_new_with_range(1, 31, 1);

	gtk_scale_set_digits(GTK_SCALE(scale), 0);

	gtk_scale_set_value_pos(GTK_SCALE(scale), GTK_POS_BOTTOM);

	if (timeout == 0)
		value = 31;
	else if (timeout < 60)
		value = 1;
	else if (timeout > 60 * 30)
		value = 30;
	else
		value = timeout / 60;

	gtk_range_set_value(GTK_RANGE(scale), value);

	gtk_range_set_update_policy(GTK_RANGE(scale), GTK_UPDATE_DISCONTINUOUS);

	gtk_box_pack_start(GTK_BOX(vbox), scale, FALSE, FALSE, 0);

	adapter->timeout_scale = scale;

	g_signal_connect(G_OBJECT(scale), "value-changed",
					G_CALLBACK(scale_callback), adapter);

	g_signal_connect(G_OBJECT(scale), "format-value",
					G_CALLBACK(format_callback), NULL);

	if (mode && strcmp(mode, "discoverable") && strcmp(mode, "limited")) {
		gtk_widget_set_sensitive(GTK_WIDGET(label), FALSE);
		gtk_widget_set_sensitive(GTK_WIDGET(scale), FALSE);
	}

	if (mode && !strcmp(mode, "limited") && adapter->enable_limited == FALSE) {
		gtk_widget_set_sensitive(GTK_WIDGET(label), FALSE);
		gtk_widget_set_sensitive(GTK_WIDGET(scale), FALSE);
	}

	label = create_label(_("\nAdapter name"));

	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	entry = gtk_entry_new();

	gtk_entry_set_max_length(GTK_ENTRY(entry), 248);

	if (name != NULL)
		gtk_entry_set_text(GTK_ENTRY(entry), name);

	gtk_box_pack_start(GTK_BOX(vbox), entry, FALSE, FALSE, 0);

	adapter->entry = entry;

	g_signal_connect(G_OBJECT(entry), "changed",
					G_CALLBACK(name_callback), adapter);

	g_signal_connect(G_OBJECT(entry), "focus-out-event",
					G_CALLBACK(focus_callback), adapter);

	label = create_label(_("\nClass of device"));

	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

	combobox = gtk_combo_box_new_text();

	gtk_combo_box_append_text(GTK_COMBO_BOX(combobox), _("Unspecified"));
	gtk_combo_box_append_text(GTK_COMBO_BOX(combobox), _("Desktop workstation"));
	gtk_combo_box_append_text(GTK_COMBO_BOX(combobox), _("Laptop computer"));

	if (major && minor && !strcmp(major, "computer")) {
		if (!strcmp(minor, "uncategorized"))
			index = 0;
		else if (!strcmp(minor, "desktop"))
			index = 1;
		else if (!strcmp(minor, "laptop"))
			index = 2;
		else
			index = -1;

		gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), index);
	} else
		gtk_widget_set_sensitive(GTK_WIDGET(combobox), FALSE);

	gtk_box_pack_start(GTK_BOX(vbox), combobox, FALSE, FALSE, 0);

	adapter->combo = combobox;

	g_signal_connect(G_OBJECT(combobox), "changed",
					G_CALLBACK(class_callback), adapter);

	gtk_widget_show_all(vbox);
}

static void mode_changed(DBusGProxy *object,
				const char *mode, gpointer user_data)
{
	GList *list;
	const char *path;

	path = dbus_g_proxy_get_path(object);

	list = g_list_find_custom(adapter_list, path, adapter_compare);
	if (list && list->data) {
		struct adapter_data *adapter = list->data;
		GtkWidget *button = NULL;
		gboolean sensitive;

		if (!adapter->attached)
			return;

		sensitive = strcmp(mode, "off") ? TRUE : FALSE;

		gtk_widget_set_sensitive(GTK_WIDGET(adapter->button_connect), sensitive);
		gtk_widget_set_sensitive(GTK_WIDGET(adapter->button_visible), sensitive);
		if (adapter->enable_limited == TRUE)
			gtk_widget_set_sensitive(GTK_WIDGET(adapter->button_limited), sensitive);
		else
			gtk_widget_set_sensitive(GTK_WIDGET(adapter->button_limited), FALSE);

		if (!strcmp(mode, "connectable")) {
			sensitive = FALSE;
			button = adapter->button_connect;
		} else if (!strcmp(mode, "discoverable")) {
			sensitive = TRUE;
			button = adapter->button_visible;
		} else if (!strcmp(mode, "limited")) {
			sensitive = adapter->enable_limited;
			button = adapter->button_limited;
		}

		gtk_widget_set_sensitive(GTK_WIDGET(adapter->timeout_label), sensitive);
		gtk_widget_set_sensitive(GTK_WIDGET(adapter->timeout_scale), sensitive);

		if (!button)
			return;

		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
	}
}

static void timeout_changed(DBusGProxy *object,
				const guint32 timeout, gpointer user_data)
{
	GList *list;
	const char *path;

	path = dbus_g_proxy_get_path(object);

	list = g_list_find_custom(adapter_list, path, adapter_compare);
	if (list && list->data) {
		struct adapter_data *adapter = list->data;
		gdouble value;

		if (timeout == 0)
			value = 31;
		else if (timeout < 60)
			value = 1;
		else if (timeout > 60 * 30)
			value = 30;
		else
			value = timeout / 60;

		gtk_range_set_value(GTK_RANGE(adapter->timeout_scale), value);
	}
}

static void name_changed(DBusGProxy *object,
				const char *name, gpointer user_data)
{
	GList *list;
	const char *path;

	path = dbus_g_proxy_get_path(object);

	list = g_list_find_custom(adapter_list, path, adapter_compare);
	if (list && list->data) {
		struct adapter_data *adapter = list->data;
		GtkWidget *child = adapter->child;
		GtkWidget *label;

		if (!adapter->attached)
			return;

		gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), child,
				name && strlen(name) > 0 ? name : _("Adapter"));

		label = gtk_notebook_get_tab_label(GTK_NOTEBOOK(notebook), child);

		gtk_label_set_max_width_chars(GTK_LABEL(label), 20);

		gtk_label_set_ellipsize(GTK_LABEL(label), PANGO_ELLIPSIZE_END);

		gtk_entry_set_text(GTK_ENTRY(adapter->entry), name ? name : "");

		adapter->name_changed = 0;
	}
}

static void minor_changed(DBusGProxy *object,
				const char *minor, gpointer user_data)
{
	GList *list;
	const char *path;

	path = dbus_g_proxy_get_path(object);

	list = g_list_find_custom(adapter_list, path, adapter_compare);
	if (list && list->data) {
		struct adapter_data *adapter = list->data;
		gint index;

		if (!adapter->attached)
			return;

		if (!strcmp(minor, "uncategorized"))
			index = 0;
		else if (!strcmp(minor, "desktop"))
			index = 1;
		else if (!strcmp(minor, "laptop"))
			index = 2;
		else
			index = -1;

		gtk_combo_box_set_active(GTK_COMBO_BOX(adapter->combo), index);
	}
}

static struct adapter_data *add_adapter(const char *path)
{
	DBusGProxy *object;
	GList *list;
	struct adapter_data *adapter;

	list = g_list_find_custom(adapter_list, path, adapter_compare);
	if (list && list->data) {
		struct adapter_data *adapter = list->data;

		adapter->attached = 1;

		create_adapter(adapter);

		return adapter;
	}

	adapter = g_try_malloc0(sizeof(*adapter));
	if (!adapter)
		return NULL;

	adapter->path = g_strdup(path);
	adapter->attached = 1;

	adapter_list = g_list_append(adapter_list, adapter);

	object = dbus_g_proxy_new_for_name(conn, "org.bluez",
						path, "org.bluez.Adapter");

	dbus_g_proxy_add_signal(object, "ModeChanged",
					G_TYPE_STRING, G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(object, "ModeChanged",
				G_CALLBACK(mode_changed), NULL, NULL);

	dbus_g_proxy_add_signal(object, "DiscoverableTimeoutChanged",
					G_TYPE_UINT, G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(object, "DiscoverableTimeoutChanged",
				G_CALLBACK(timeout_changed), NULL, NULL);

	dbus_g_proxy_add_signal(object, "NameChanged",
					G_TYPE_STRING, G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(object, "NameChanged",
				G_CALLBACK(name_changed), NULL, NULL);

	dbus_g_proxy_add_signal(object, "MinorClassChanged",
					G_TYPE_STRING, G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(object, "MinorClassChanged",
				G_CALLBACK(minor_changed), NULL, NULL);

	create_adapter(adapter);

	return adapter;
}

static void adapter_added(DBusGProxy *object,
				const char *path, gpointer user_data)
{
	struct adapter_data *adapter;
	GtkTreeModel *model;

	adapter = add_adapter(path);

	model = bluetooth_client_get_model(client);

	gtk_tree_view_set_model(GTK_TREE_VIEW(device_tree), model);

	g_object_unref(model);

	gtk_tree_view_expand_all(GTK_TREE_VIEW(device_tree));
}

static void adapter_removed(DBusGProxy *object,
				const char *path, gpointer user_data)
{
	GList *list;

	list = g_list_find_custom(adapter_list, path, adapter_compare);
	if (list && list->data) {
		struct adapter_data *adapter = list->data;
		GtkTreeModel *model;
		gint page;

		adapter->attached = 0;

		model = bluetooth_client_get_model(client);

		gtk_tree_view_set_model(GTK_TREE_VIEW(device_tree), model);

		g_object_unref(model);

		page = gtk_notebook_page_num(GTK_NOTEBOOK(notebook),
							adapter->child);

		gtk_notebook_remove_page(GTK_NOTEBOOK(notebook), page);

		adapter->child = NULL;
	}
}

static GList *service_list = NULL;

struct service_data {
	char *path;
	int present;
	GtkTreeIter iter;
};

static void service_free(gpointer data, gpointer user_data)
{
	struct service_data *service = data;

	service_list = g_list_remove(service_list, service);

	g_free(service->path);
	g_free(service);
}

static void service_disable(gpointer data, gpointer user_data)
{
	struct service_data *service = data;

	if (service->present) {
		service->present = 0;
		gtk_list_store_remove(service_store, &service->iter);
	}
}

static gint service_compare(gconstpointer a, gconstpointer b)
{
	const struct service_data *service = a;
	const char *path = b;

	return strcmp(service->path, path);
}

static void service_count(gpointer data, gpointer user_data)
{
	struct service_data *service = data;
	int *count = user_data;

	if (service->present)
		(*count)++;
}

static int present_services(void)
{
	int count = 0;

	g_list_foreach(service_list, service_count, &count);

	return count;
}

static void service_started(DBusGProxy *object, gpointer user_data)
{
	struct service_data *service = user_data;

	gtk_list_store_set(service_store, &service->iter, 2, TRUE, -1);
}

static void service_stopped(DBusGProxy *object, gpointer user_data)
{
	struct service_data *service = user_data;

	gtk_list_store_set(service_store, &service->iter, 2, FALSE, -1);
}

static void add_service(const char *path)
{
	DBusGProxy *object;
	GList *list;
	struct service_data *service;
	const char *name = NULL;
	const char *description = NULL;
	gboolean running;

	list = g_list_find_custom(service_list, path, service_compare);
	if (list && list->data) {
		service = list->data;
		goto done;
	}

	service = g_try_malloc0(sizeof(*service));
	if (!service)
		return;

	service->path = g_strdup(path);

	service_list = g_list_append(service_list, service);

done:
	service->present = 1;

	object = dbus_g_proxy_new_for_name(conn, "org.bluez",
						path, "org.bluez.Service");

	dbus_g_proxy_add_signal(object, "Started", G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(object, "Started",
				G_CALLBACK(service_started), service, NULL);

	dbus_g_proxy_add_signal(object, "Stopped", G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(object, "Stopped",
				G_CALLBACK(service_stopped), service, NULL);

	dbus_g_proxy_call(object, "GetName", NULL, G_TYPE_INVALID,
				G_TYPE_STRING, &name, G_TYPE_INVALID);

	dbus_g_proxy_call(object, "GetDescription", NULL, G_TYPE_INVALID,
				G_TYPE_STRING, &description, G_TYPE_INVALID);

	dbus_g_proxy_call(object, "IsRunning", NULL, G_TYPE_INVALID,
				G_TYPE_BOOLEAN, &running, G_TYPE_INVALID);

	gtk_list_store_insert_with_values(service_store,
				&service->iter, -1, 0, path,
				1, name ? name : path, 2, running,
				4, description ? description : name, -1);
}

static void service_added(DBusGProxy *object,
				const char *path, gpointer user_data)
{
	add_service(path);
}

static void service_removed(DBusGProxy *object,
				const char *path, gpointer user_data)
{
	GList *list;

	list = g_list_find_custom(service_list, path, service_compare);
	if (list && list->data) {
		struct service_data *service = list->data;

		service->present = 0;

		gtk_list_store_remove(service_store, &service->iter);

		if (present_services() > 0)
			return;

		gtk_widget_set_sensitive(GTK_WIDGET(button_properties), FALSE);
	}
}

static DBusGProxy *setup_manager(void)
{
	DBusGProxy *object;
	GError *error = NULL;
	const gchar **array = NULL;

	object = dbus_g_proxy_new_for_name(conn, "org.bluez",
					"/org/bluez", "org.bluez.Manager");
	if (!object)
		return NULL;

	dbus_g_proxy_add_signal(object, "AdapterAdded",
					G_TYPE_STRING, G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(object, "AdapterAdded",
				G_CALLBACK(adapter_added), conn, NULL);

	dbus_g_proxy_add_signal(object, "AdapterRemoved",
					G_TYPE_STRING, G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(object, "AdapterRemoved",
				G_CALLBACK(adapter_removed), conn, NULL);

	dbus_g_proxy_call(object, "ListAdapters", &error,
			G_TYPE_INVALID, G_TYPE_STRV, &array, G_TYPE_INVALID);

	if (error == NULL) {
		while (*array) {
			add_adapter(*array);
			array++;
		}
	} else {
		g_error_free(error);
		error = NULL;
	}

	service_store = gtk_list_store_new(5, G_TYPE_STRING, G_TYPE_STRING,
				G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_STRING);
	if (!service_store) {
		g_object_unref(object);
		return NULL;
	}

	dbus_g_proxy_add_signal(object, "ServiceAdded",
					G_TYPE_STRING, G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(object, "ServiceAdded",
				G_CALLBACK(service_added), conn, NULL);

	dbus_g_proxy_add_signal(object, "ServiceRemoved",
					G_TYPE_STRING, G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(object, "ServiceRemoved",
				G_CALLBACK(service_removed), conn, NULL);

	dbus_g_proxy_call(object, "ListServices", &error,
			G_TYPE_INVALID, G_TYPE_STRV, &array, G_TYPE_INVALID);

	if (error == NULL) {
		while (*array) {
			add_service(*array);
			array++;
		}
	} else
		g_error_free(error);

	return object;
}

static void name_owner_changed(DBusGProxy *object, const char *name,
			const char *prev, const char *new, gpointer user_data)
{
	if (!strcmp(name, "org.bluez") && *new == '\0') {
		g_list_foreach(adapter_list, adapter_disable, NULL);
		g_list_foreach(service_list, service_disable, NULL);
		gtk_widget_set_sensitive(GTK_WIDGET(button_properties), FALSE);
	}
}

static DBusGProxy *setup_dbus(void)
{
	DBusGProxy *object;

	object = dbus_g_proxy_new_for_name(conn, DBUS_SERVICE_DBUS,
					DBUS_PATH_DBUS, DBUS_INTERFACE_DBUS);

	dbus_g_proxy_add_signal(object, "NameOwnerChanged",
		G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(object, "NameOwnerChanged",
				G_CALLBACK(name_owner_changed), NULL, NULL);

	return object;
}

static void gconf_callback(GConfClient *client, guint cnxn_id,
					GConfEntry *entry, gpointer user_data)
{
	if (gconf_entry_get_value(entry) == NULL)
		return;

	if (strcmp(entry->key, PREF_USE_HAL) == 0) {
		gboolean value;

		value = gconf_client_get_bool(client, entry->key, NULL);

		if (!button_usehal)
			return;

		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button_usehal), value);
	}

	if (strcmp(entry->key, PREF_ICON_POLICY) == 0) {
		char *str;

		str = gconf_client_get_string(client, entry->key, NULL);
		if (str) {
			gconf_string_to_enum(icon_policy_enum_map,
							str, &icon_policy);
			g_free(str);
		}

		update_icon_policy(NULL);
	}

	if (strcmp(entry->key, PREF_AUTO_AUTHORIZE) == 0) {
		gboolean value;

		value = gconf_client_get_bool(client, entry->key, NULL);

		if (!button_autoauth)
			return;

		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button_autoauth), value);
	}
}

int main(int argc, char *argv[])
{
	DBusGProxy *service, *manager;
	GError *error = NULL;
	char *str;

	bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
	bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
	textdomain(GETTEXT_PACKAGE);

	gtk_init(&argc, &argv);

	conn = dbus_g_bus_get(DBUS_BUS_SYSTEM, &error);
	if (error != NULL) {
		g_printerr(_("Connecting to system bus failed: %s\n"),
							error->message);
		g_error_free(error);
		exit(EXIT_FAILURE);
	}

	gconf = gconf_client_get_default();

	str = gconf_client_get_string(gconf, PREF_ICON_POLICY, NULL);
	if (str) {
		gconf_string_to_enum(icon_policy_enum_map, str, &icon_policy);
		g_free(str);
	}

	gconf_client_add_dir(gconf, PREF_DIR, GCONF_CLIENT_PRELOAD_NONE, NULL);

	gconf_client_notify_add(gconf, PREF_DIR,
					gconf_callback, NULL, NULL, NULL);

	client = bluetooth_client_new();

	notebook = gtk_notebook_new();

	service = setup_dbus();

	manager = setup_manager();

	create_window();

	gtk_main();

	g_object_unref(gconf);

	g_list_foreach(service_list, service_free, NULL);

	g_list_foreach(adapter_list, adapter_free, NULL);

	g_object_unref(manager);

	g_object_unref(service);

	dbus_g_connection_unref(conn);

	g_object_unref(client);

	return 0;
}
